/*
 * Linux-specific portion of
 * Broadcom 802.11abg Networking Device Driver
 *
 * Copyright 2005-2006, Broadcom Corporation
 * All Rights Reserved.                
 *                                     
 * THIS SOFTWARE IS OFFERED "AS IS", AND BROADCOM GRANTS NO WARRANTIES OF ANY
 * KIND, EXPRESS OR IMPLIED, BY STATUTE, COMMUNICATION OR OTHERWISE. BROADCOM
 * SPECIFICALLY DISCLAIMS ANY IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS
 * FOR A SPECIFIC PURPOSE OR NONINFRINGEMENT CONCERNING THIS SOFTWARE.
 *
 * $Id$
 */

#define __UNDEF_NO_VERSION__

#include <wlc_cfg.h>

#include <typedefs.h>

#include <linux/module.h>
#include <linuxver.h>
#include <osl.h>

#include <linux/types.h>
#include <linux/errno.h>
#include <linux/pci.h>
#include <linux/init.h>
#include <linux/kernel.h>
#include <linux/proc_fs.h>
#include <linux/netdevice.h>
#include <linux/etherdevice.h>
#include <linux/skbuff.h>
#include <linux/delay.h>
#include <linux/string.h>
#include <linux/ethtool.h>
#include <linux/completion.h>

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 14)
#include <net/ieee80211.h>
#endif

#include <asm/system.h>
#include <asm/io.h>
#include <asm/irq.h>
#include <asm/pgtable.h>
#include <asm/uaccess.h>
#include <asm/unaligned.h>

#include <epivers.h>
#include <bcmendian.h>
#include <proto/ethernet.h>
#include <bcmdevs.h>
#include <bcmdefs.h>
#include <bcmutils.h>
#include <pcicfg.h>
#include <wlioctl.h>
#include <wl_linux.h>
#include <wlc_key.h>

typedef void wlc_info_t;
typedef const struct sb_pub  sb_t;
#include <wlc_pub.h>
#include <wl_dbg.h>
#include <wlc_ethereal.h>

/* Linux wireless extension support */
#ifdef CONFIG_NET_RADIO
#include <wl_iw.h>
#endif /* CONFIG_NET_RADIO */

#include <wl_export.h>

typedef struct wl_timer {
	struct timer_list timer;
	struct wl_info *wl;
	void (*fn)(void *);
	uint ms;
	bool periodic;
	bool set;
	struct wl_timer *next;
} wl_timer_t;

/* contortion to call functions at safe time */
typedef struct wl_task {
	struct work_struct work;
	void *context;
} wl_task_t;

#define WL_IFTYPE_BSS	1 /* iftype subunit for BSS */

typedef struct wl_if {
	struct wl_if *next;
	struct wl_info *wl;		/* back pointer to main wl_info_t */
	struct net_device *dev;		/* virtual netdevice */
	int type;			/* interface type: BSS */
	struct wlc_if *wlcif;		/* wlc interface handle */
	struct ether_addr remote;	/* remote partner */
	uint subunit;			/* BSS unit */
} wl_if_t;

struct wl_info {
#ifdef CONFIG_NET_RADIO
	wl_iw_t		iw;		/* wireless extensions state (must be first) */
#endif /* CONFIG_NET_RADIO */
	wlc_pub_t	*pub;		/* pointer to public wlc state */
	void		*wlc;		/* pointer to private common os-independent data */
	osl_t		*osh;		/* pointer to os handler */
	struct net_device *dev;		/* backpoint to device */
	spinlock_t	lock;		/* per-device perimeter lock */
	uint		bustype;	/* bus type */
	bool		piomode;	/* set from insmod argument */
	void *regsva;			/* opaque chip registers virtual address */
	struct net_device_stats stats;	/* stat counter reporting structure */
	wl_if_t *if_list;		/* list of all interfaces */
	struct wl_info *next;		/* pointer to next wl_info_t in chain */
	volatile uint callbacks;	/* # outstanding callback functions */
	struct wl_timer *timers;	/* timer cleanup queue */
	struct tasklet_struct tasklet;	/* dpc tasklet */
	ulong flags;			/* current irq flags */
	struct net_device *monitor;	/* monitor pseudo device */
	bool		resched;	/* dpc needs to be and is rescheduled */

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 14)
	struct ieee80211_crypto_ops *tkipmodops;	/* external tkip module ops */
	struct ieee80211_tkip_data  *tkip_ucast_data;
	struct ieee80211_tkip_data  *tkip_bcast_data;
#endif
};

static int wl_found = 0;

/* defines */
#define	DEV_WLPTR(dev)		dev->priv	/* priv points to wl */
#define	DEV_WLIFPTR(dev)	dev->dn_ptr	
#define	WL_INFO(dev)	(wl_info_t*)(DEV_WLPTR(dev))
#define	WL_DEV_IF(dev)	((wl_if_t*)(DEV_WLIFPTR(dev)))
#define WL_LOCK(wl)	do { ulong flags; spin_lock_irqsave(&(wl)->lock, flags); \
	(wl)->flags = flags; } while (0)
#define WL_UNLOCK(wl)	do { ulong flags; flags = (wl)->flags; \
	spin_unlock_irqrestore(&(wl)->lock, flags); } while (0)
#define WL_LOCKED(wl)	spin_is_locked(&(wl)->lock)

#define WL_TASKLET (LINUX_VERSION_CODE > KERNEL_VERSION(2, 4, 5))

/* local prototypes */
static int wl_open(struct net_device *dev);
static int wl_close(struct net_device *dev);
static int wl_start(struct sk_buff *skb, struct net_device *dev);
static struct net_device_stats *wl_get_stats(struct net_device *dev);
#ifdef CONFIG_NET_RADIO
static struct iw_statistics *wl_get_wireless_stats(struct net_device *dev);
#endif /* CONFIG_NET_RADIO */
static int wl_set_mac_address(struct net_device *dev, void *addr);
static void wl_set_multicast_list(struct net_device *dev);
#if LINUX_VERSION_CODE > KERNEL_VERSION(2, 4, 2)
static int wl_ethtool(wl_info_t *wl, void *uaddr);
#endif /* LINUX_VERSION_CODE > KERNEL_VERSION(2, 4, 2) */
#if WL_TASKLET
static void wl_dpc(ulong data);
#endif	/* WL_TASKLET */
static void wl_link_up(wl_info_t *wl);
static void wl_link_down(wl_info_t *wl);

static void wl_tkipmic_error(wl_info_t *wl, struct ether_addr *ea, bool group, bool flush_txq);

static int wl_schedule_task(wl_info_t *wl, void (*fn)(struct wl_task *), void *context);
#if defined(CONFIG_PROC_FS) && (LINUX_VERSION_CODE >= KERNEL_VERSION(2, 4, 0))
static int wl_read_proc(char *buffer, char **start, off_t offset, int length, int *eof, void *data);
#endif /* defined(CONFIG_PROC_FS) && (LINUX_VERSION_CODE >= KERNEL_VERSION(2, 4, 0)) */
struct wl_if *wl_alloc_if(wl_info_t *wl, int iftype, uint unit, struct wlc_if* wlc_if);
static void wl_free_if(wl_info_t *wl, wl_if_t *wlif);


#ifdef CONFIG_PCI
/* recognized PCI IDs */
static struct pci_device_id wl_id_table[] __devinitdata = {
	{ vendor: PCI_ANY_ID,
	device: PCI_ANY_ID,
	subvendor: PCI_ANY_ID,
	subdevice: PCI_ANY_ID,
	class: PCI_CLASS_NETWORK_OTHER << 8,
	class_mask: 0xffff00,
	driver_data: 0,
	},
	{ 0, }
};
MODULE_DEVICE_TABLE(pci, wl_id_table);
#endif	/* CONFIG_PCI */


static int oneonly = 0;
module_param(oneonly, int, 0);

static int piomode = 0;
module_param(piomode, int, 0);

static char name[IFNAMSIZ] = "eth%d";
module_param_string(name, name, IFNAMSIZ, 0);

/* BCMSLTGT: slow target */

#ifndef	SRCBASE
#define	SRCBASE "."
#endif /* SRCBASE */

/** 
 * attach to the WL device.
 *
 * Attach to the WL device identified by vendor and device parameters.
 * regs is a host accessible memory address pointing to WL device registers.
 *
 * wl_attach is not defined as static because in the case where no bus
 * is defined, wl_attach will never be called, and thus, gcc will issue
 * a warning that this function is defined but not used if we declare
 * it as static.
 */
wl_info_t *
wl_attach(uint16 vendor, uint16 device, ulong regs, uint bustype, void *btparam, uint irq)
{
	struct net_device *dev;
	wl_if_t *wlif;
	wl_info_t *wl;
#if defined(CONFIG_PROC_FS) && (LINUX_VERSION_CODE >= KERNEL_VERSION(2, 4, 0))
	char tmp[128];
#endif
	osl_t *osh;
	int unit;
	uint err;

	unit = wl_found;

	if (oneonly && (unit != 0)) {
		WL_ERROR(("wl%d: wl_attach: oneonly is set, exiting\n", unit));
		return NULL;
	}

	/* Requires pkttag feature */
	osh = osl_attach(btparam, TRUE);
	ASSERT(osh);

	/* allocate private info */
	if ((wl = (wl_info_t*) MALLOC(osh, sizeof(wl_info_t))) == NULL) {
		WL_ERROR(("wl%d: malloc wl_info_t, out of memory, malloced %d bytes\n", unit,
			MALLOCED(osh)));
		osl_detach(osh);
		return NULL;
	}
	bzero(wl, sizeof(wl_info_t));

	wl->osh = osh;

	wlif = wl_alloc_if(wl, WL_IFTYPE_BSS, unit, NULL);
	if (!wlif) {
		WL_ERROR(("wl%d: wl_alloc_if failed\n", unit));
		MFREE(osh, wl, sizeof(wl_info_t));
		osl_detach(osh);
		return NULL;
	}

	dev = wlif->dev;
	strncpy(dev->name, name, IFNAMSIZ);

	wl->dev = dev;

	/* map chip registers (47xx: and sprom) */
	dev->base_addr = regs;

	if (bustype == PCI_BUS) {
		/* piomode can be overwritten by command argument */
		wl->piomode = piomode;
		WL_TRACE(("PCI/%s\n", wl->piomode ? "PIO" : "DMA"));
	}
	else {
		WL_ERROR(("wl%d: Only PCI BUS devices are supported \n", unit));
		goto fail;
	}
	wl->bustype = bustype;

	if ((wl->regsva = ioremap_nocache(dev->base_addr, PCI_BAR0_WINSZ)) == NULL) {
		WL_ERROR(("wl%d: ioremap() failed\n", unit));
		goto fail;
	}

	spin_lock_init(&wl->lock);

	/* common load-time initialization */
	if (!(wl->wlc = wlc_attach((void *) wl, vendor, device, unit, wl->piomode,
		osh, wl->regsva, wl->bustype, btparam, &err))) {
		printf("%s: %s driver failed with code %d\n", dev->name, EPI_VERSION_STR, err);
		goto fail;
	}
	wl->pub = (wlc_pub_t *)wl->wlc;

#if defined(CONFIG_PROC_FS) && (LINUX_VERSION_CODE >= KERNEL_VERSION(2, 4, 0))
	/* create /proc/net/wl<unit> */
	sprintf(tmp, "net/wl%d", wl->pub->unit);
	create_proc_read_entry(tmp, 0, 0, wl_read_proc, (void*)wl);
#endif /* defined(CONFIG_PROC_FS) && (LINUX_VERSION_CODE >= KERNEL_VERSION(2, 4, 0)) */

	bcopy(&wl->pub->cur_etheraddr, dev->dev_addr, ETHER_ADDR_LEN);

#if WL_TASKLET
	/* setup the bottom half handler */
	tasklet_init(&wl->tasklet, wl_dpc, (ulong)wl);
#endif /* WL_TASKLET */

	/* register our interrupt handler */
	{
		if (request_irq(irq, wl_isr, SA_SHIRQ, dev->name, wl)) {
			WL_ERROR(("wl%d: request_irq() failed\n", unit));
			goto fail;
		}
		dev->irq = irq;
	}

	/* lastly, enable our entry points */
	dev->open = wl_open;
	dev->stop = wl_close;
	dev->hard_start_xmit = wl_start;
	dev->get_stats = wl_get_stats;
	dev->set_mac_address = wl_set_mac_address;
	dev->set_multicast_list = wl_set_multicast_list;
	dev->do_ioctl = wl_ioctl;
#ifdef CONFIG_NET_RADIO
	dev->get_wireless_stats = wl_get_wireless_stats;
#if WIRELESS_EXT > 12
	dev->wireless_handlers = (struct iw_handler_def *) &wl_iw_handler_def;
#endif /* WIRELESS_EXT > 12 */
#endif /* CONFIG_NET_RADIO */

	if (register_netdev(dev)) {
		WL_ERROR(("wl%d: register_netdev() failed\n", unit));
		goto fail;
	}

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 14)
	/* load the tkip module */
	wl->tkipmodops = ieee80211_get_crypto_ops("TKIP");
	if (wl->tkipmodops == NULL) {
		request_module("ieee80211_crypt_tkip");
		wl->tkipmodops = ieee80211_get_crypto_ops("TKIP");
	}
	WL_ERROR(("TkipModule OPs at 0x%x\n",  (uint)wl->tkipmodops));
#endif
	wl->iw.wlinfo = (void *)wl;

	/* print hello string */
	printf("%s: Broadcom BCM%04x 802.11 Wireless Controller " EPI_VERSION_STR,
		dev->name, device);
	printf("\n");

	wl_found++;
	return wl;

fail:
	wl_free(wl);
	return NULL;
}

#if defined(CONFIG_PROC_FS) && (LINUX_VERSION_CODE >= KERNEL_VERSION(2, 4, 0))
static int
wl_read_proc(char *buffer, char **start, off_t offset, int length, int *eof, void *data)
{
	wl_info_t *wl;
	int len;
	off_t pos;
	off_t begin;

	len = pos = begin = 0;

	wl = (wl_info_t*) data;

	len = wl_dump(wl, buffer, WLC_IOCTL_MAXLEN);
	pos = begin + len;

	if (pos < offset) {
		len = 0;
		begin = pos;
	}

	*eof = 1;

	*start = buffer + (offset - begin);
	len -= (offset - begin);

	if (len > length)
		len = length;

	return (len);
}
#endif /* defined(CONFIG_PROC_FS) && (LINUX_VERSION_CODE >= KERNEL_VERSION(2, 4, 0)) */

#ifdef CONFIG_PCI
/** 
 * determines if a device is a WL device, and if so, attaches it.
 *
 * This function determines if a device pointed to by pdev is a WL device,
 * and if so, performs a wl_attach() on it.
 *
 */
int __devinit
wl_pci_probe(struct pci_dev *pdev, const struct pci_device_id *ent)
{
	wl_info_t *wl;

	WL_TRACE(("wl_pci_probe: bus %d slot %d func %d irq %d\n",
		pdev->bus->number, PCI_SLOT(pdev->devfn), PCI_FUNC(pdev->devfn), pdev->irq));

	if (!wlc_chipmatch(pdev->vendor, pdev->device))
		return (-ENODEV);

	pci_set_master(pdev);
	pci_enable_device(pdev);

	wl = wl_attach(pdev->vendor, pdev->device, pci_resource_start(pdev, 0), PCI_BUS, pdev,
		pdev->irq);
	if (!wl)
		return -ENODEV;

	pci_set_drvdata(pdev, wl);

	return 0;
}

static int
wl_suspend(struct pci_dev *pdev, u32 state)
{
	wl_info_t *wl;

	WL_TRACE(("wl: wl_suspend\n"));

	if ((wl = (wl_info_t *) pci_get_drvdata(pdev))) {
		netif_device_detach(wl->dev);
		WL_LOCK(wl);
		wl_down(wl);
		WL_UNLOCK(wl);
	}

	return 0;
}

static int
wl_resume(struct pci_dev *pdev)
{
	wl_info_t *wl;
	int error;

	error = 0;

	if ((wl = (wl_info_t *) pci_get_drvdata(pdev))) {
		WL_LOCK(wl);
		error = wl_up(wl);
		WL_UNLOCK(wl);
		netif_device_attach(wl->dev);
	}

	return (error);
}

/* Compatibility routines */
#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 4, 6)
static void
_wl_suspend(struct pci_dev *pdev)
{
	wl_suspend(pdev, 0);
}

static void
_wl_resume(struct pci_dev *pdev)
{
	wl_resume(pdev);
}
#endif /* LINUX_VERSION_CODE < KERNEL_VERSION(2, 4, 6) */

static void __devexit
wl_remove(struct pci_dev *pdev)
{
	wl_info_t *wl;

	if (!wlc_chipmatch(pdev->vendor, pdev->device)) {
		return;
	}

	if ((wl = (wl_info_t *) pci_get_drvdata(pdev))) {
		wl_suspend(pdev, 0);
		wl_free(wl);
		pci_set_drvdata(pdev, NULL);
	}
}

static struct pci_driver wl_pci_driver = {
	name:		"wl",
	probe:		wl_pci_probe,
#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 4, 6)
	suspend:	_wl_suspend,
	resume:		_wl_resume,
#else
	suspend:	wl_suspend,
	resume:		wl_resume,
#endif /* LINUX_VERSION_CODE < KERNEL_VERSION(2, 4, 6) */
	remove:		__devexit_p(wl_remove),
	id_table:	wl_id_table,
	};
#endif	/* CONFIG_PCI */




/** 
 * This is the main entry point for the WL driver.
 *
 * This function determines if a device pointed to by pdev is a WL device,
 * and if so, performs a wl_attach() on it.
 *
 */
static int __init
wl_module_init(void)
{
	int error = -ENODEV;




#ifdef CONFIG_PCI
	if (!(error = pci_module_init(&wl_pci_driver)))
		return (0);
#endif	/* CONFIG_PCI */
	return (error);
}

/** 
 * This function unloads the WL driver from the system.
 *
 * This function unconditionally unloads the WL driver module from the
 * system.
 *
 */
static void __exit
wl_module_exit(void)
{


#ifdef CONFIG_PCI
	pci_unregister_driver(&wl_pci_driver);
#endif	/* CONFIG_PCI */
}

module_init(wl_module_init);
module_exit(wl_module_exit);

/** 
 * This function frees the WL per-device resources.
 *
 * This function frees resources owned by the WL device pointed to
 * by the wl parameter.
 *
 */
void
wl_free(wl_info_t *wl)
{
	wl_timer_t *t, *next;
	osl_t *osh;

	WL_TRACE(("wl: wl_free\n"));
	{
		if (wl->dev && wl->dev->irq)
			free_irq(wl->dev->irq, wl);
	}

	if (wl->dev) {
		wl_free_if(wl, WL_DEV_IF(wl->dev));
		wl->dev = NULL;
	}

#if WL_TASKLET
	/* kill dpc */
	tasklet_kill(&wl->tasklet);
#endif /* WL_TASKLET */

	/* free common resources */
	if (wl->wlc) {
#if defined(CONFIG_PROC_FS) && (LINUX_VERSION_CODE >= KERNEL_VERSION(2, 4, 0))
		char tmp[128];
		/* remove /proc/net/wl<unit> */
		sprintf(tmp, "net/wl%d", wl->pub->unit);
		remove_proc_entry(tmp, 0);
#endif /* defined(CONFIG_PROC_FS) && (LINUX_VERSION_CODE >= KERNEL_VERSION(2, 4, 0)) */
		wlc_detach(wl->wlc);
		wl->wlc = NULL;
		wl->pub = NULL;
	}

	/* virtual interface deletion is deferred so we cannot spinwait */

	/* wait for all pending callbacks to complete */
	while (wl->callbacks > 0)
		schedule();

	/* free timers */
	for (t = wl->timers; t; t = next) {
		next = t->next;
		MFREE(wl->osh, t, sizeof(wl_timer_t));
	}

	/* free monitor */
	if (wl->monitor) {
		unregister_netdev(wl->monitor);
		MFREE(wl->osh, wl->monitor, sizeof(struct net_device));
		wl->monitor = NULL;
	}

	osh = wl->osh;

	/*
	 * unregister_netdev() calls get_stats() which may read chip registers
	 * so we cannot unmap the chip registers until after calling unregister_netdev() .
	 */
	if (wl->regsva)
		iounmap((void*)wl->regsva);
	wl->regsva = NULL;
	
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 14)
	/* un register the TKIP module...if any */
	if (wl->tkipmodops != NULL) {
		if (wl->tkip_ucast_data) {
			wl->tkipmodops->deinit(wl->tkip_ucast_data);
			wl->tkip_ucast_data = NULL;
		}
		if (wl->tkip_bcast_data) {
			wl->tkipmodops->deinit(wl->tkip_bcast_data);
			wl->tkip_bcast_data = NULL;
		}
	}
#endif

	MFREE(osh, wl, sizeof(wl_info_t));

	if (MALLOCED(osh)) {
		printf("Memory leak of bytes %d\n", MALLOCED(osh));
		ASSERT(0);
	}

	osl_detach(osh);
}


static int
wl_open(struct net_device *dev)
{
	wl_info_t *wl;
	int error;

	wl = WL_INFO(dev);

	WL_TRACE(("wl%d: wl_open\n", wl->pub->unit));

	WL_LOCK(wl);
	error = wl_up(wl);
	if (!error) {
		WL_ERROR(("about to set the promisc bit %d\n", error));
		error = wlc_set(wl->wlc, WLC_SET_PROMISC, (dev->flags & IFF_PROMISC));
		WL_ERROR(("after the set to the promisc bit %d\n", error));
	}
	WL_UNLOCK(wl);

	if (!error)
		OLD_MOD_INC_USE_COUNT;

	return (error? -ENODEV: 0);
}

static int
wl_close(struct net_device *dev)
{
	wl_info_t *wl;

	wl = WL_INFO(dev);

	WL_TRACE(("wl%d: wl_close\n", wl->pub->unit));

	WL_LOCK(wl);
	wl_down(wl);
	WL_UNLOCK(wl);

	OLD_MOD_DEC_USE_COUNT;

	return (0);
}

/* transmit a packet */
static int
wl_start(struct sk_buff *skb, struct net_device *dev)
{
	wl_info_t *wl;
	wl_if_t *wlif;
	void *pkt;

	wlif = WL_DEV_IF(dev);
	wl = wlif->wl;

	WL_TRACE(("wl%d: wl_start: len %d\n", wl->pub->unit, skb->len));

	/* Convert the packet. Mainly attach a pkttag */
	if ((pkt = PKTFRMNATIVE(wl->osh, skb)) == NULL) {
		dev_kfree_skb_any(skb);
		return -ENOMEM;
	}

	WL_LOCK(wl);

	wlc_sendpkt(wl->wlc, pkt, wlif->wlcif);

	WL_UNLOCK(wl);

	return (0);
}

void
wl_txflowcontrol(wl_info_t *wl, bool state)
{
	wl_if_t *wlif;

	for (wlif = wl->if_list; wlif != NULL; wlif = wlif->next) {
		if (state == ON)
			netif_stop_queue(wlif->dev);
		else
			netif_wake_queue(wlif->dev);
	}
}

/* Schedule a completion handler to run at safe time */
static int
wl_schedule_task(wl_info_t *wl, void (*fn)(struct wl_task *task), void *context)
{
	wl_task_t *task;

	WL_TRACE(("wl%d: wl_schedule_task\n", wl->pub->unit));

	if (!(task = MALLOC(wl->osh, sizeof(wl_task_t)))) {
		WL_ERROR(("wl%d: wl_schedule_task: out of memory, malloced %d bytes\n",
			wl->pub->unit, MALLOCED(wl->osh)));
		return -ENOMEM;
	}

	INIT_WORK(&task->work, (void (*)(void *)) fn, task);
	task->context = context;

	if (!schedule_work(&task->work)) {
		WL_ERROR(("wl%d: schedule_work() failed\n", wl->pub->unit));
		MFREE(wl->osh, task, sizeof(wl_task_t));
		return -ENOMEM;
	}

	wl->callbacks++;
	return 0;
}

struct wl_if *
wl_alloc_if(wl_info_t *wl, int iftype, uint subunit, struct wlc_if* wlcif)
{
	struct net_device *dev;
	wl_if_t *wlif;
	wl_if_t *p;

	if (!(wlif = MALLOC(wl->osh, sizeof(wl_if_t) + sizeof(struct net_device)))) {
		WL_ERROR(("wl%d: wl_alloc_if: out of memory, malloced %d bytes\n",
		          (wl->pub)?wl->pub->unit:subunit, MALLOCED(wl->osh)));
		return NULL;
	}
	bzero(wlif, sizeof(wl_if_t) + sizeof(struct net_device));

	dev = (struct net_device *) &wlif[1];
	wlif->type = iftype;
	wlif->dev = dev;
	wlif->wl = wl;
	wlif->wlcif = wlcif;
	wlif->subunit = subunit;

	ether_setup(dev);
	DEV_WLPTR(dev) = wl;
	DEV_WLIFPTR(dev) = wlif;

	/* match current flow control state */
	if (wl->dev && netif_queue_stopped(wl->dev))
		netif_stop_queue(dev);

	/* add the interface to the interface linked list */
	if (wl->if_list == NULL)
		wl->if_list = wlif;
	else {
		p = wl->if_list;
		while (p->next != NULL)
			p = p->next;
		p->next = wlif;
	}

	return wlif;
}

static void
wl_free_if(wl_info_t *wl, wl_if_t *wlif)
{
	wl_if_t *p;

	/* remove the interface from the interface linked list */
	p = wl->if_list;
	if (p == wlif)
		wl->if_list = p->next;
	else {
		while (p != NULL && p->next != wlif)
			p = p->next;
		if (p != NULL)
			p->next = p->next->next;
	}

	DEV_WLIFPTR(wlif->dev) = NULL;

	unregister_netdev(wlif->dev);
	MFREE(wl->osh, wlif, sizeof(wl_if_t) + sizeof(struct net_device));
}


/* Return pointer to interface name */
char *
wl_ifname(wl_info_t *wl, wl_if_t *wlif)
{
	if (wlif)
		return wlif->dev->name;
	else
		return wl->dev->name;
}

void
wl_init(wl_info_t *wl)
{
	WL_TRACE(("wl%d: wl_init\n", wl->pub->unit));

	wl_reset(wl);

	wlc_init(wl->wlc);
}

uint
wl_reset(wl_info_t *wl)
{
	WL_TRACE(("wl%d: wl_reset\n", wl->pub->unit));

	wlc_reset(wl->wlc);

	return (0);
}

/*
 * These are interrupt on/off entry points.
 * Since the wl_isr is serialized with other driver entries using a spinlock,
 * they are SMP safe, just call common routine directly,
 */
void
wl_intrson(wl_info_t *wl)
{
	wlc_intrson(wl->wlc);
}

uint32
wl_intrsoff(wl_info_t *wl)
{
	return wlc_intrsoff(wl->wlc);
}

void
wl_intrsrestore(wl_info_t *wl, uint32 macintmask)
{
	wlc_intrsrestore(wl->wlc, macintmask);
}

int
wl_up(wl_info_t *wl)
{
	int error = 0;

	WL_TRACE(("wl%d: wl_up\n", wl->pub->unit));

	if (wl->pub->up)
		return (0);

	error = wlc_up(wl->wlc);

	/* wake (not just start) all interfaces */
	if (!error)
		wl_txflowcontrol(wl, OFF);

	return (error);
}

void
wl_down(wl_info_t *wl)
{
	wl_if_t *wlif;
	uint callbacks, ret_val;

	WL_TRACE(("wl%d: wl_down\n", wl->pub->unit));

	for (wlif = wl->if_list; wlif != NULL; wlif = wlif->next) {
		netif_down(wlif->dev);
		netif_stop_queue(wlif->dev);
	}

	/* call common down function */
	ret_val  = wlc_down(wl->wlc);
	callbacks = wl->callbacks - ret_val;

	/* wait for down callbacks to complete */
	WL_UNLOCK(wl);

	SPINWAIT((wl->callbacks > callbacks), 100 * 1000);
	ASSERT(wl->callbacks == callbacks);

	WL_LOCK(wl);
}

#if LINUX_VERSION_CODE > KERNEL_VERSION(2, 4, 2)
static int
wl_ethtool(wl_info_t *wl, void *uaddr)
{
	struct ethtool_drvinfo info;
	uint32 cmd;

	if (copy_from_user(&cmd, uaddr, sizeof(uint32)))
		return (-EFAULT);

	switch (cmd) {
	case ETHTOOL_GDRVINFO:
		bzero(&info, sizeof(info));
		info.cmd = cmd;
		sprintf(info.driver, "wl%d", wl->pub->unit);
		strcpy(info.version, EPI_VERSION_STR);
		if (copy_to_user(uaddr, &info, sizeof(info)))
			return (-EFAULT);
		return (0);
	}

	return (-EOPNOTSUPP);
}
#endif /* LINUX_VERSION_CODE > KERNEL_VERSION(2, 4, 2) */

int
wl_ioctl(struct net_device *dev, struct ifreq *ifr, int cmd)
{
	wl_info_t *wl;
	wl_if_t *wlif;
	void *buf = NULL;
	wl_ioctl_t ioc;
	int bcmerror;
	int status;

	wl = WL_INFO(dev);
	wlif = WL_DEV_IF(dev);

	bcmerror = 0;
	status = 0;

	WL_TRACE(("wl%d: wl_ioctl: cmd 0x%x\n", wl->pub->unit, cmd));

#ifdef CONFIG_NET_RADIO
	/* linux wireless extensions */
	if ((cmd >= SIOCIWFIRST) && (cmd <= SIOCIWLAST)) {
		/* may recurse, do NOT lock */
		return wl_iw_ioctl(dev, ifr, cmd);
	}
#endif /* CONFIG_NET_RADIO */

#if LINUX_VERSION_CODE > KERNEL_VERSION(2, 4, 2)
	if (cmd == SIOCETHTOOL)
		return (wl_ethtool(wl, (void*)ifr->ifr_data));
#endif /* LINUX_VERSION_CODE > KERNEL_VERSION(2, 4, 2) */

	if (cmd != SIOCDEVPRIVATE) {
		bcmerror = BCME_UNSUPPORTED;
		goto done2;
	}

	if (copy_from_user(&ioc, ifr->ifr_data, sizeof(wl_ioctl_t))) {
		bcmerror = BCME_BADADDR;
		goto done2;
	}

	/* optimization for direct ioctl calls from kernel */
	if (segment_eq(get_fs(), KERNEL_DS))
		buf = ioc.buf;

	else if (ioc.buf) {
		if (!(buf = (void *) MALLOC(wl->osh, MAX(ioc.len, WLC_IOCTL_MAXLEN)))) {
			bcmerror = BCME_NORESOURCE;
			goto done2;
		}

		if (copy_from_user(buf, ioc.buf, ioc.len)) {
			bcmerror = BCME_BADADDR;
			goto done1;
		}
	}

	WL_LOCK(wl);
	if (!capable(CAP_NET_ADMIN))
		bcmerror = BCME_EPERM;
	else {
		bcmerror = wlc_ioctl(wl->wlc, ioc.cmd, buf, ioc.len, wlif->wlcif);
	}
	WL_UNLOCK(wl);

done1:
	if (ioc.buf && (ioc.buf != buf)) {
		if (copy_to_user(ioc.buf, buf, ioc.len))
			bcmerror = BCME_BADADDR;
		MFREE(wl->osh, buf, MAX(ioc.len, WLC_IOCTL_MAXLEN));
	}

	if (status)
		return status;
done2:
	ASSERT(VALID_BCMERROR(bcmerror));
	if (bcmerror != 0)
		wl->pub->bcmerror = bcmerror;
	return (OSL_ERROR(bcmerror));
}

static struct net_device_stats*
wl_get_stats(struct net_device *dev)
{
	struct net_device_stats *stats;
	wl_info_t *wl = WL_INFO(dev);

	WL_TRACE(("wl%d: wl_get_stats\n", wl->pub->unit));

	stats = &wl->stats;

	WL_LOCK(wl);

	/* refresh stats */
	if (wl->pub->up)
		wlc_statsupd(wl->wlc);

	stats->rx_packets = WLCNTVAL(wl->pub->_cnt.rxframe);
	stats->tx_packets = WLCNTVAL(wl->pub->_cnt.txframe);
	stats->rx_bytes = WLCNTVAL(wl->pub->_cnt.rxbyte);
	stats->tx_bytes = WLCNTVAL(wl->pub->_cnt.txbyte);
	stats->rx_errors = WLCNTVAL(wl->pub->_cnt.rxerror);
	stats->tx_errors = WLCNTVAL(wl->pub->_cnt.txerror);
	stats->collisions = 0;

	stats->rx_length_errors = 0;
	stats->rx_over_errors = WLCNTVAL(wl->pub->_cnt.rxoflo);
	stats->rx_crc_errors = WLCNTVAL(wl->pub->_cnt.rxcrc);
	stats->rx_frame_errors = 0;
	stats->rx_fifo_errors = WLCNTVAL(wl->pub->_cnt.rxoflo);
	stats->rx_missed_errors = 0;

	stats->tx_fifo_errors = WLCNTVAL(wl->pub->_cnt.txuflo);

	WL_UNLOCK(wl);

	return (stats);
}

#ifdef CONFIG_NET_RADIO
static struct iw_statistics *
wl_get_wireless_stats(struct net_device *dev)
{
	int res;
	wl_info_t *wl;
	wl_if_t *wlif;
	struct iw_statistics *wstats;
	int phy_noise, rssi;

	wl = WL_INFO(dev);
	wlif = WL_DEV_IF(dev);
	wstats = &wl->iw.wstats;

	WL_TRACE(("wl%d: wl_get_wireless_stats\n", wl->pub->unit));

	WL_LOCK(wl);

	if ((res = wlc_get(wl->wlc, WLC_GET_PHY_NOISE, &phy_noise)))
		goto done;

	{
		scb_val_t scb;
		if ((res = wlc_ioctl(wl->wlc, WLC_GET_RSSI, &scb, sizeof(scb), wlif->wlcif)))
			goto done;
		rssi = scb.val;
	}

	if (rssi <= WLC_RSSI_NO_SIGNAL)
		wstats->qual.qual = 0;
	else if (rssi <= WLC_RSSI_VERY_LOW)
		wstats->qual.qual = 1;
	else if (rssi <= WLC_RSSI_LOW)
		wstats->qual.qual = 2;
	else if (rssi <= WLC_RSSI_GOOD)
		wstats->qual.qual = 3;
	else if (rssi <= WLC_RSSI_VERY_GOOD)
		wstats->qual.qual = 4;
	else
		wstats->qual.qual = 5;

	/* Wraps to 0 if RSSI is 0 */
	wstats->qual.level = 0x100 + rssi;
	wstats->qual.noise = 0x100 + phy_noise;
	wstats->qual.updated |= 7;

#if WIRELESS_EXT > 11
	wstats->discard.nwid = 0; 
	wstats->discard.code = WLCNTVAL(wl->pub->_cnt.rxundec);
	wstats->discard.fragment = WLCNTVAL(wl->pub->_cnt.rxfragerr);
	wstats->discard.retries = WLCNTVAL(wl->pub->_cnt.txfail);
	wstats->discard.misc = WLCNTVAL(wl->pub->_cnt.rxrunt) + WLCNTVAL(wl->pub->_cnt.rxgiant);

	wstats->miss.beacon = 0; 
#endif /* WIRELESS_EXT > 11 */

done:
	WL_UNLOCK(wl);

	if (res == 0)
		return wstats;
	else
		return NULL;
}
#endif /* CONFIG_NET_RADIO */

static int
wl_set_mac_address(struct net_device *dev, void *addr)
{
	wl_info_t *wl;
	struct sockaddr *sa = (struct sockaddr *) addr;

	wl = WL_INFO(dev);

	WL_TRACE(("wl%d: wl_set_mac_address\n", wl->pub->unit));

	bcopy(sa->sa_data, dev->dev_addr, ETHER_ADDR_LEN);
	if (wlc_iovar_op(wl->wlc, "cur_etheraddr", NULL, 0, sa->sa_data, ETHER_ADDR_LEN,
		IOV_SET, (WL_DEV_IF(dev))->wlcif))
		WL_ERROR(("wl%d: wl_set_mac_address: error setting MAC addr override\n",
			wl->pub->unit));

	return 0;
}

static void
wl_set_multicast_list(struct net_device *dev)
{
	wl_info_t *wl;
	struct dev_mc_list *mclist;
	int i;

	wl = WL_INFO(dev);

	WL_TRACE(("wl%d: wl_set_multicast_list\n", wl->pub->unit));

	WL_LOCK(wl);

	if (wl->pub->up) {
		wl->pub->allmulti = (dev->flags & IFF_ALLMULTI)? TRUE: FALSE;

		/* copy the list of multicasts into our private table */
		for (i = 0, mclist = dev->mc_list; mclist && (i < dev->mc_count);
			i++, mclist = mclist->next) {
			if (i >= MAXMULTILIST) {
				wl->pub->allmulti = TRUE;
				i = 0;
				break;
			}
			wl->pub->multicast[i] = *((struct ether_addr*) mclist->dmi_addr);
		}
		wl->pub->nmulticast = i;
		wlc_set(wl->wlc, WLC_SET_PROMISC, (dev->flags & IFF_PROMISC));
	}

	WL_UNLOCK(wl);
}

irqreturn_t
wl_isr(int irq, void *dev_id, struct pt_regs *ptregs)
{
	wl_info_t *wl;
	bool ours, wantdpc;

	wl = (wl_info_t*) dev_id;

	spin_lock(&wl->lock);

	/* call common first level interrupt handler */
	if ((ours = wlc_isr(wl->wlc, &wantdpc))) {
		/* if more to do... */
		if (wantdpc) {
			/* ...disable interrupts... */
			wlc_intrsoff_isr(wl->wlc);

			/* ...and call the second level interrupt handler */
#if WL_TASKLET
			/* schedule dpc */
			ASSERT(wl->resched == FALSE);
			tasklet_schedule(&wl->tasklet);
#else
			wlc_dpc(wl->wlc, FALSE);
			wlc_intrson(wl->wlc);
#endif /* WL_TASKLET */
		}
	}

	spin_unlock(&wl->lock);

	return IRQ_RETVAL(ours);
}

#if WL_TASKLET
static void
wl_dpc(ulong data)
{
	wl_info_t *wl;

	wl = (wl_info_t*) data;

	WL_LOCK(wl);

	/* call the common second level interrupt handler */
	if (wl->pub->up) {
		if (wl->resched)
			wlc_intrsupd(wl->wlc);
		wl->resched = wlc_dpc(wl->wlc, TRUE);
	}

	/* wlc_dpc() may bring the driver down */
	if (!wl->pub->up)
		goto done;

	/* re-schedule dpc */
	if (wl->resched)
		tasklet_schedule(&wl->tasklet);
	/* re-enable interrupts */
	else
		wlc_intrson(wl->wlc);

done:
	WL_UNLOCK(wl);
}
#endif	/* WL_TASKLET */

void
wl_sendup(wl_info_t *wl, wl_if_t *wlif, void *p)
{
	struct sk_buff *skb;

	WL_TRACE(("wl%d: wl_sendup: %d bytes\n", wl->pub->unit, PKTLEN(wl->osh, p)));

	/* route packet to the appropriate interface */
	if (wlif) {
		/* drop if the interface is not up yet */
		if (!netif_device_present(wlif->dev)) {
			WL_ERROR(("wl%d: wl_sendup: interface not ready\n", wl->pub->unit));
			PKTFREE(wl->osh, p, FALSE);
			return;
		}
		/* Convert the packet, mainly detach the pkttag */
		skb = PKTTONATIVE(wl->osh, p);
		skb->dev = wlif->dev;
	} else {
		/* Convert the packet, mainly detach the pkttag */
		skb = PKTTONATIVE(wl->osh, p);
		skb->dev = wl->dev;
	}

	skb->protocol = eth_type_trans(skb, wl->dev);

	ASSERT(ISALIGNED((uintptr)skb->data, 4));

	/* send it up */
	netif_rx(skb);
}

int
wl_dump(wl_info_t *wl, char *buf, uint len)
{
	/* big enough? */
	if (len < WLC_IOCTL_MAXLEN)
		return (WLC_IOCTL_MAXLEN);

	sprintf(buf, "wl%d: %s %s version %s\n", wl->pub->unit,
		__DATE__, __TIME__, EPI_VERSION_STR);

	wlc_tinydump(wl->wlc, &buf[strlen(buf)]);

	ASSERT(strlen(buf) < len);

	return (strlen(buf));
}


static void
wl_link_up(wl_info_t *wl)
{
	WL_ERROR(("wl%d: link up\n", wl->pub->unit));
}

static void
wl_link_down(wl_info_t *wl)
{
	WL_ERROR(("wl%d: link down\n", wl->pub->unit));
}

void
wl_event(wl_info_t *wl, wlc_event_t *e)
{
#ifdef CONFIG_NET_RADIO
	wl_iw_event(wl->dev, e);
#endif /* CONFIG_NET_RADIO */

	switch (e->event.event_type) {
	case WLC_E_LINK:
	case WLC_E_NDIS_LINK:
		if (e->event.flags&WLC_EVENT_MSG_LINK)
			wl_link_up(wl);
		else
			wl_link_down(wl);
		break;
	case WLC_E_MIC_ERROR:
		wl_tkipmic_error(wl, e->addr,
			e->event.flags&WLC_EVENT_MSG_GROUP,
			e->event.flags&WLC_EVENT_MSG_FLUSHTXQ);
		break;
	}
}

static void
wl_timer(ulong data)
{
	wl_timer_t *t;

	t = (wl_timer_t*) data;

	WL_LOCK(t->wl);

	if (t->set) {
		if (t->periodic) {
			t->timer.expires = jiffies + t->ms*HZ/1000;
			t->wl->callbacks++;
			add_timer(&t->timer);
			t->set = TRUE;
		} else
			t->set = FALSE;


		t->fn(t->wl->wlc);
	}

	t->wl->callbacks--;

	WL_UNLOCK(t->wl);
}

wl_timer_t *
wl_init_timer(wl_info_t *wl, void (*fn)(void *wlc))
{
	wl_timer_t *t;

	if (!(t = MALLOC(wl->osh, sizeof(wl_timer_t)))) {
		WL_ERROR(("wl%d: wl_init_timer: out of memory, malloced %d bytes\n", wl->pub->unit,
			MALLOCED(wl->osh)));
		return 0;
	}

	init_timer(&t->timer);
	t->timer.data = (ulong) t;
	t->timer.function = wl_timer;
	t->wl = wl;
	t->fn = fn;
	t->set = FALSE;
	t->periodic = FALSE;
	t->ms = 0;
	t->next = wl->timers;
	wl->timers = t;

	return t;
}

void
wl_add_timer(wl_info_t *wl, wl_timer_t *t, uint ms, int periodic)
{
	ASSERT(!t->set);

	t->ms = ms;
	t->periodic = (bool) periodic;
	t->set = TRUE;
	t->timer.expires = jiffies + ms*HZ/1000;

	wl->callbacks++;
	add_timer(&t->timer);
}

/* return TRUE if timer successfully deleted, FALSE if still pending */
bool
wl_del_timer(wl_info_t *wl, wl_timer_t *t)
{
	if (t->set) {
		t->set = FALSE;
		if (!del_timer(&t->timer))
			return FALSE;
		wl->callbacks--;
	}

	return TRUE;
}

void
wl_free_timer(wl_info_t *wl, wl_timer_t *t)
{
	ASSERT(!t->set);
}

static void
wl_tkipmic_error(wl_info_t *wl, struct ether_addr *ea, bool group, bool flush_txq)
{
	WL_WSEC(("wl%d: mic error using %s key\n", wl->pub->unit,
		(group) ? "group" : "pairwise"));


}

void
wl_monitor(wl_info_t *wl, wl_rxsts_t *rxsts, void *p)
{
	struct sk_buff *oskb = (struct sk_buff *)p;
	p80211msg_t *phdr;
	struct sk_buff *skb;
	uint len;
	uchar *pdata;

	WL_TRACE(("wl%d: wl_monitor\n", wl->pub->unit));

	if (!wl->monitor)
		return;

	len = sizeof(p80211msg_t) + oskb->len - D11_PHY_HDR_LEN;
	if ((skb = dev_alloc_skb(len)) == NULL)
		return;

	skb_put(skb, len);
	phdr = (p80211msg_t*)skb->data;
	/* Initialize the message members */
	phdr->msgcode = WL_MON_FRAME;
	phdr->msglen = sizeof(p80211msg_t);
	strcpy(phdr->devname, wl->dev->name);

	phdr->hosttime.did = WL_MON_FRAME_HOSTTIME;
	phdr->hosttime.status = P80211ITEM_OK;
	phdr->hosttime.len = 4;
	phdr->hosttime.data = jiffies;

	phdr->channel.did = WL_MON_FRAME_CHANNEL;
	phdr->channel.status = P80211ITEM_NO_VALUE;
	phdr->channel.len = 4;
	phdr->channel.data = 0;

	phdr->signal.did = WL_MON_FRAME_SIGNAL;
	phdr->signal.status = P80211ITEM_NO_VALUE;
	phdr->signal.len = 4;
	phdr->signal.data = (rxsts->preamble == WL_RXS_PREAMBLE_SHORT) ? 1 : 0;

	phdr->noise.did = WL_MON_FRAME_NOISE;
	phdr->noise.status = P80211ITEM_NO_VALUE;
	phdr->noise.len = 4;
	phdr->noise.data = 0;

	phdr->rate.did = WL_MON_FRAME_RATE;
	phdr->rate.status = P80211ITEM_OK;
	phdr->rate.len = 4;
	phdr->rate.data = rxsts->datarate;

	phdr->istx.did = WL_MON_FRAME_ISTX;
	phdr->istx.status = P80211ITEM_NO_VALUE;
	phdr->istx.len = 4;
	phdr->istx.data = 0;

	phdr->mactime.did = WL_MON_FRAME_MACTIME;
	phdr->mactime.status = P80211ITEM_OK;
	phdr->mactime.len = 4;
	phdr->mactime.data = rxsts->mactime;

	phdr->rssi.did = WL_MON_FRAME_RSSI;
	phdr->rssi.status = P80211ITEM_OK;
	phdr->rssi.len = 4;
	phdr->rssi.data = rxsts->signal;		/* to dbm */

	phdr->sq.did = WL_MON_FRAME_SQ;
	phdr->sq.status = P80211ITEM_OK;
	phdr->sq.len = 4;
	phdr->sq.data = rxsts->sq;

	phdr->frmlen.did = WL_MON_FRAME_FRMLEN;
	phdr->frmlen.status = P80211ITEM_OK;
	phdr->frmlen.status = P80211ITEM_OK;
	phdr->frmlen.len = 4;
	phdr->frmlen.data = rxsts->pktlength;

	pdata = skb->data + sizeof(p80211msg_t);
	bcopy(oskb->data + D11_PHY_HDR_LEN, pdata, oskb->len - D11_PHY_HDR_LEN);

	skb->dev = wl->monitor;
	skb->dev->last_rx = jiffies;
	skb->mac.raw = skb->data;
	skb->ip_summed = CHECKSUM_NONE;
	skb->pkt_type = PACKET_OTHERHOST;
	skb->protocol = htons(ETH_P_80211_RAW);

	netif_rx(skb);
}

static int
wl_monitor_start(struct sk_buff *skb, struct net_device *dev)
{
	dev_kfree_skb(skb);
	return 0;
}

static void
wl_add_monitor(wl_task_t *task)
{
	wl_info_t *wl = (wl_info_t *) task->context;

	WL_LOCK(wl);

	if (wl->monitor)
		goto done;

	if (!(wl->monitor = MALLOC(wl->osh, sizeof(struct net_device))))
		goto done;
	bzero(wl->monitor, sizeof(struct net_device));

	DEV_WLPTR(wl->monitor) = wl;
	wl->monitor->hard_start_xmit = wl_monitor_start;
	wl->monitor->do_ioctl = wl_ioctl;
	sprintf(wl->monitor->name, "prism%d", wl->pub->unit);
	bcopy(wl->dev->dev_addr, wl->monitor->dev_addr, ETHER_ADDR_LEN);
	wl->monitor->type = ARPHRD_IEEE80211_PRISM;
	wl->monitor->flags = wl->dev->flags;

	if (register_netdev(wl->monitor)) {
		MFREE(wl->osh, wl->monitor, sizeof(struct net_device));
		wl->monitor = NULL;
		goto done;
	}

done:
	WL_ERROR(("wl%d: wl_add_monitor: out of memory, malloced %d bytes\n", wl->pub->unit,
		MALLOCED(wl->osh)));
	MFREE(wl->osh, task, sizeof(wl_task_t));
	wl->callbacks--;
	WL_UNLOCK(wl);
}

static void
wl_del_monitor(wl_task_t *task)
{
	wl_info_t *wl = (wl_info_t *) task->context;

	WL_LOCK(wl);

	if (!wl->monitor)
		goto done;

	unregister_netdev(wl->monitor);
	MFREE(wl->osh, wl->monitor, sizeof(struct net_device));
	wl->monitor = NULL;

done:
	MFREE(wl->osh, task, sizeof(wl_task_t));
	wl->callbacks--;
	WL_UNLOCK(wl);
}

/*
 * Create a dedicated monitor interface since libpcap caches the
 * packet type when it opens the device. The protocol type in the skb
 * is dropped somewhere in libpcap, and every received frame is tagged
 * with the DLT/ARPHRD type that's read by libpcap when the device is
 * opened.
 *
 * If libpcap was fixed to handle per-packet link types, we might not
 * need to create a pseudo device at all, wl_set_monitor() would be
 * unnecessary, and wlc->monitor could just get set in wlc_ioctl().
 */
void
wl_set_monitor(wl_info_t *wl, int val)
{
	WL_TRACE(("wl%d: wl_set_monitor: val %d\n", wl->pub->unit, val));

	if (val && !wl->monitor)
		(void) wl_schedule_task(wl, wl_add_monitor, wl);
	else if (!val && wl->monitor)
		(void) wl_schedule_task(wl, wl_del_monitor, wl);
}

struct net_device *
wl_netdev_get(wl_info_t *wl)
{
	return wl->dev;
}

int
wl_tkip_miccheck(wl_info_t *wl, void *p, int hdr_len, bool group_key, int key_index)
{

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 14)
	struct sk_buff *skb;
	skb = PKTTONATIVE(wl->osh, p);
	if (wl->tkipmodops) {
		if (group_key && wl->tkip_bcast_data)
			return(wl->tkipmodops->decrypt_msdu(skb, key_index, hdr_len, wl->tkip_bcast_data));
		else if (!group_key && wl->tkip_ucast_data)
			return(wl->tkipmodops->decrypt_msdu(skb, key_index, hdr_len, wl->tkip_ucast_data));
	}
#endif
	/* Error */
	WL_ERROR(("%s: No tkip mod ops\n", __FUNCTION__));
	return -1;

}

int
wl_tkip_micadd(wl_info_t *wl, void *p, int hdr_len)
{
	int error = -1;

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 14)
	struct sk_buff *skb;
	skb = PKTTONATIVE(wl->osh, p);
	if (wl->tkipmodops) {
		if (wl->tkip_ucast_data)
			error = wl->tkipmodops->encrypt_msdu(skb, hdr_len, wl->tkip_ucast_data);
		if (error)
			WL_ERROR(("Error encrypting MSDU %d\n", error));
	}
	else 
#endif
		WL_ERROR(("%s: No tkip mod ops\n", __FUNCTION__));
	return error;
	
}

int
wl_tkip_encrypt(wl_info_t *wl, void *p, int hdr_len)
{
	int error = -1;
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 14)
	struct sk_buff *skb;
	skb = PKTTONATIVE(wl->osh, p);
	if (wl->tkipmodops) {
		if (wl->tkip_ucast_data)
			error = wl->tkipmodops->encrypt_mpdu(skb, hdr_len, wl->tkip_ucast_data);
		if (error) {
			WL_ERROR(("Error encrypting MPDU %d\n", error));
		}
	}
	else 
#endif
		WL_ERROR(("%s: No tkip mod ops\n", __FUNCTION__));
	return error;
	
}

int
wl_tkip_decrypt(wl_info_t *wl, void *p, int hdr_len, bool group_key)
{
	int err = -1;
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 14)
	struct sk_buff *skb;

	skb = PKTTONATIVE(wl->osh, p);
	if (wl->tkipmodops) {
		if (group_key && wl->tkip_bcast_data)
			err = wl->tkipmodops->decrypt_mpdu(skb, hdr_len, wl->tkip_bcast_data);
		else if (!group_key && wl->tkip_ucast_data)
			err = wl->tkipmodops->decrypt_mpdu(skb, hdr_len, wl->tkip_ucast_data);
	}
	else 
		WL_ERROR(("%s: No tkip mod ops\n", __FUNCTION__));

#endif 
		
	/* Error */
	return err;
}

int
wl_tkip_keyset(wl_info_t *wl, wsec_key_t *key)
{
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 14)
	bool group_key = FALSE;	
	uchar	rxseq[IW_ENCODE_SEQ_MAX_SIZE];

	if (key->len != 0) {
		WL_WSEC(("%s: Key Length is Not zero\n", __FUNCTION__));
		if (key->algo != CRYPTO_ALGO_TKIP) {
			WL_WSEC(("%s: Algo is Not TKIP %d\n", __FUNCTION__, key->algo));
			return 0;
		}
		WL_WSEC(("%s: Trying to set a key in TKIP Mod\n", __FUNCTION__));
	}
	else 
		WL_WSEC(("%s: Trying to Remove a Key from TKIP Mod\n", __FUNCTION__));

	if (ETHER_ISNULLADDR(&key->ea) || ETHER_ISBCAST(&key->ea)) {
		group_key = TRUE;
		WL_WSEC(("Group Key index %d\n", key->id));
	}
	else 
		WL_WSEC(("Unicast Key index %d\n", key->id));

	if (wl->tkipmodops) {
		if (group_key) {
			if (key->len) {
				if (!wl->tkip_bcast_data) {
					WL_WSEC(("Init TKIP Bcast Module\n"));
					WL_UNLOCK(wl);
					wl->tkip_bcast_data = wl->tkipmodops->init(key->id);
					WL_LOCK(wl);
				}
				if (wl->tkip_bcast_data) {
					bzero(rxseq, IW_ENCODE_SEQ_MAX_SIZE);
					bcopy(&key->rxiv, rxseq, 6);
					wl->tkipmodops->set_key(&key->data, key->len, (uint8 *)&key->rxiv, wl->tkip_bcast_data);
				}
			}
			else {
				if (wl->tkip_bcast_data) {
					WL_WSEC(("Deinit TKIP Bcast Module\n"));
					wl->tkipmodops->deinit(wl->tkip_bcast_data);
					wl->tkip_bcast_data = NULL;
				}
			}
		}
		else {
			if (key->len) {
				if (!wl->tkip_ucast_data) {
					WL_WSEC(("Init TKIP Ucast Module\n"));
					WL_UNLOCK(wl);
					wl->tkip_ucast_data = wl->tkipmodops->init(key->id);
					WL_LOCK(wl);
				}
				if (wl->tkip_ucast_data) {
					bzero(rxseq, IW_ENCODE_SEQ_MAX_SIZE);
					bcopy(&key->rxiv, rxseq, 6);
					wl->tkipmodops->set_key(&key->data, key->len, (uint8 *)&key->rxiv, wl->tkip_ucast_data);
				}
			}
			else {
				if (wl->tkip_ucast_data) {
					WL_WSEC(("Deinit TKIP Ucast Module\n"));
					wl->tkipmodops->deinit(wl->tkip_ucast_data);
					wl->tkip_ucast_data = NULL;
				}
			}
		}
	}
	else 
#endif
		WL_WSEC(("%s: No tkip mod ops\n", __FUNCTION__));
	return 0;
}

void 
wl_tkip_printstats(wl_info_t *wl, bool group_key)
{
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 14)
	char debug_buf[512];
	if (wl->tkipmodops) {
		if (group_key && wl->tkip_bcast_data)
			wl->tkipmodops->print_stats(debug_buf, wl->tkip_bcast_data);
		else if (!group_key && wl->tkip_ucast_data)
			wl->tkipmodops->print_stats(debug_buf, wl->tkip_ucast_data);
		else 
			return;
		printk("%s: TKIP stats from module: %s\n", debug_buf, group_key?"Bcast":"Ucast");
	}
#endif
}
